#!/usr/bin/env python
# vim: et ai ts=4 sw=4:

from charmhelpers.core import hookenv
from charmhelpers.core.hookenv import ERROR, INFO

import json
import os
import sys

hooks = hookenv.Hooks()

PERSIST_DATA_FILE = "charm-data"


def _persist_data(key, value):
    """Save C{key} and C{value} pairs in a persistent file for reference"""
    data = {}
    filepath = "%s/%s" % (hookenv.charm_dir(), PERSIST_DATA_FILE)
    if os.path.exists(filepath):
        with open(filepath) as inputfile:
            data = json.load(inputfile)
    data[key] = value
    with open(filepath, "w") as outfile:
        outfile.write(unicode(json.dumps(data, ensure_ascii=False)))


def _get_persistent_data(key=None, remove_values=False):
    """Get a persistent data as specified by C{key} from the PERSIST_DATA_FILE
    If C{remove_values} is C{True} remove the specified C{key} from the data
    file.
    """
    data = {}
    filepath = "%s/%s" % (hookenv.charm_dir(), PERSIST_DATA_FILE)
    with open(filepath) as inputfile:
        data = json.load(inputfile)
    if key:
        value = data.get(key, None)
        if remove_values:
            data.pop(key, None)
            if data:
                with open(filepath, "w") as outfile:
                    outfile.write(
                        unicode(json.dumps(data, ensure_ascii=False)))
    else:
        value = data
        if remove_values:
            if os.path.exists(filepath):
                os.remove(filepath)
    return value


@hooks.hook()
def config_changed():
    """Update changed endpoints and credentials"""
    import nova_util
    missing_options = []
    for config_option in REQUIRED_CONFIGURATION_OPTIONS:
        if not hookenv.config(config_option):
            missing_options.append(config_option)
    if missing_options:
        hookenv.log(
            "WARNING: Incomplete charm config. Missing values for: %s" %
            ",".join(missing_options), hookenv.WARNING)
        sys.exit(0)
    nova_util.load_environment()


@hooks.hook()
def install():
    """Install required packages if not present"""
    from charmhelpers import fetch
    required_packages = ["python-novaclient"]

    fetch.add_source("cloud-archive:havana")
    fetch.apt_update(fatal=True)
    fetch.apt_install(required_packages, fatal=True)


@hooks.hook("block-storage-relation-departed")
def block_storage_relation_departed():
    """Detach a nova volume from the related instance when relation departs"""
    import nova_util
    # XXX juju bug:1279018 for departed-hooks to see relation-data
    instance_id = _get_persistent_data(
        hookenv.remote_unit(), remove_values=True)
    if not instance_id:
        hookenv.log(
            "Cannot detach nova volume from instance without instance-id",
            ERROR)
        sys.exit(1)
    nova_util.detach_nova_volume(instance_id)


@hooks.hook("block-storage-relation-changed")
def block_storage_relation_changed():
    """Attach a nova device to the C{instance-id} requested by the relation

    Optionally the relation can specify:
        - C{volume-label} the label to set on the created volume
        - C{volume-id} to attach an existing volume-id or volume-name
        - C{size} to request a specific volume size
    """
    import nova_util
    instance_id = hookenv.relation_get('instance-id')
    volume_label = hookenv.relation_get('volume-label')
    _persist_data(hookenv.remote_unit(), instance_id)
    if not instance_id:
        hookenv.log("Waiting for relation to define instance-id", INFO)
        return
    volume_id = hookenv.relation_get('volume-id')  # optional
    size = hookenv.relation_get('size')            # optional
    device_path = nova_util.attach_nova_volume(
        instance_id=instance_id, volume_id=volume_id, size=size,
        volume_label=volume_label)

    # Volume is attached, send the path back to the remote storage unit
    hookenv.relation_set(relation_settings={"block-device-path": device_path})


###############################################################################
# Global variables
###############################################################################
REQUIRED_CONFIGURATION_OPTIONS = [
    "endpoint", "region", "tenant", "key", "secret"]

if __name__ == '__main__':
    hook_name = os.path.basename(sys.argv[0])
    hookenv.log("Running {} hook".format(hook_name))
    if hookenv.relation_id():
        hookenv.log("Relation {} with {}".format(
            hookenv.relation_id(), hookenv.remote_unit()))
    hooks.execute(sys.argv)
