"""Common python utilities for the nova provider"""

from charmhelpers.core import hookenv
import subprocess
import os
import sys
from time import sleep
from hooks import REQUIRED_CONFIGURATION_OPTIONS

METADATA_URL = "http://169.254.169.254/openstack/2012-08-10/meta_data.json"
NOVA_ENVIRONMENT_MAP = {
    "endpoint": "OS_AUTH_URL", "region": "OS_REGION_NAME",
    "tenant": "OS_TENANT_NAME", "key": "OS_USERNAME", "secret": "OS_PASSWORD"}


def load_environment():
    """
    Source our credentials from the configuration definitions into our
    operating environment
    """
    config_data = hookenv.config()
    for option in REQUIRED_CONFIGURATION_OPTIONS:
        environment_variable = NOVA_ENVIRONMENT_MAP[option]
        os.environ[environment_variable] = config_data[option].strip()
    validate_credentials()


def validate_credentials():
    """Attempt to contact nova volume service or exit(1)"""
    try:
        subprocess.check_call("nova list", shell=True)
    except subprocess.CalledProcessError, e:
        hookenv.log(
            "ERROR: Charm configured credentials can't access endpoint. %s" %
            str(e),
            hookenv.ERROR)
        sys.exit(1)
    hookenv.log(
        "Validated charm configuration credentials have access to block "
        "storage service")


def get_volume_attachments(volume_id):
    """Return a C{list} of volume attachments if present"""
    from ast import literal_eval
    try:
        output = subprocess.check_output(
            "nova volume-show %s | grep attachments | awk -F '|' '{print $3}'"
            % volume_id, shell=True)
        attachments = literal_eval(output.strip())
    except subprocess.CalledProcessError:
        return []
    return attachments


def volume_exists(volume_id):
    """Returns C{True} when C{volume_id} already exists"""
    try:
        subprocess.check_call(
            "nova volume-show %s" % volume_id, shell=True)
    except subprocess.CalledProcessError:
        return False
    return True


def get_volume_id(volume_designation=None, instance_id=None):
    """Return the nova volume id associated with this unit

    Optionally, C{volume_designation} can be either a volume-id or
    volume-display-name and the matching C{volume-id} will be returned.
    If no matching volume is found, return C{None}.
    """
    token = volume_designation
    if token:
        # Get volume by name or volume-id
        # nova volume-show will error if multiple matches are found
        command = (
            "nova volume-show '%s' | grep ' id ' | awk '{ print $4 }'" % token)
    elif instance_id:
        command = "nova volume-list | grep %s | awk '{print $2}'" % instance_id
    else:
        # Find volume by unit name
        token = hookenv.remote_unit()
        command = "nova volume-list | grep %s | awk '{print $2}'" % token

    try:
        output = subprocess.check_output(command, shell=True)
    except subprocess.CalledProcessError, e:
        hookenv.log(
            "ERROR: Couldn't find nova volume id for %s. %s" % (token, str(e)),
            hookenv.ERROR)
        sys.exit(1)

    lines = output.strip().split("\n")
    if len(lines) > 1:
        hookenv.log(
            "Error: Multiple nova volumes labeled as associated with "
            "%s. Cannot get_volume_id." % token, hookenv.ERROR)
        sys.exit(1)
    if lines[0]:
        return lines[0]
    return None


def get_volume_status(volume_designation=None):
    """Return the status of a nova volume
    If C{volume_designation} is specified, return the status of that volume,
    otherwise use L{get_volume_id} to grab the volume currently related to
    this unit. If no volume is discoverable, return C{None}.
    """
    if volume_designation is None:
        volume_designation = get_volume_id()
        if volume_designation is None:
            hookenv.log(
                "WARNING: Can't find volume_id to get status.",
                hookenv.WARNING)
            return None
    try:
        output = subprocess.check_output(
            "nova volume-show '%s' | grep ' status ' | awk '{ print $4 }'"
            % volume_designation, shell=True)
    except subprocess.CalledProcessError, e:
        hookenv.log(
            "Error: nova couldn't get status of volume %s. %s" %
            (volume_designation, str(e)), hookenv.ERROR)
        return None
    return output.strip()


def attach_nova_volume(instance_id, volume_id=None, size=None,
                       volume_label=None):
    """
    Run nova commands to create and attach a volume to the remote unit if none
    exists. Attempt to attach and validate the attached volume 10 times. If
    unable to resolve the attach issues, exit in error and log the issue.

    Log errors if the nova volume is in an unsupported state, and if C{in-use}
    report it is already attached.
    Return the device-path of the attached volume to the caller.
    """
    load_environment()   # Will fail if proper environment is not set up
    remote_unit = hookenv.remote_unit()
    if volume_label is None:
        volume_label = "%s unit volume" % remote_unit
    if volume_id:
        if not volume_exists(volume_id):
            hookenv.log(
                "Requested volume-id (%s) does not exist. Unable to associate "
                "storage with %s" % (volume_id, remote_unit),
                hookenv.ERROR)
            sys.exit(1)

        # Validate that current volume status is supported
        status = get_volume_status(volume_id)
        if status in ["in-use", "attaching"]:
            hookenv.log("Volume %s already attached. Done" % volume_id)
            attachment = get_volume_attachments(volume_id)[0]
            return attachment["device"]  # The device path on the instance
        if status != "available":
            hookenv.log(
                "Cannot attach nova volume. Volume has unsupported status: %s"
                % status)
            sys.exit(1)
    else:
        # No volume_id, create a new volume if one isn't already created for
        # this JUJU_REMOTE_UNIT
        volume_id = get_volume_id(volume_label)
        if not volume_id:
            if not size:
                size = hookenv.config("default_volume_size")
            hookenv.log(
                "Creating a %sGig volume named (%s) for instance %s" %
                (size, volume_label, instance_id))
            subprocess.check_call(
                "nova volume-create --display-name '%s' %s" %
                (volume_label, size), shell=True)
            # Get new volume_id search for remote_unit in volume label
            volume_id = get_volume_id(volume_label)

    device = None
    hookenv.log("Attaching %s (%s)" % (volume_label, volume_id))
    for x in range(10):
        status = get_volume_status(volume_id)
        if status == "in-use":
            attachment = get_volume_attachments(volume_id)[0]
            return attachment["device"]  # The device path on the instance
        if status == "available":
            device = subprocess.check_output(
                "nova volume-attach %s %s auto | egrep -o \"/dev/vd[b-z]\"" %
                (instance_id, volume_id), shell=True)
            break
        else:
            sleep(5)

    if not device:
        hookenv.log(
            "ERROR: Unable to discover device attached by nova volume-attach",
            hookenv.ERROR)
        sys.exit(1)
    return device.strip()


def detach_nova_volume(instance_id):
    """Use nova commands to detach a volume from remote unit if present"""
    load_environment()   # Will fail if proper environment is not set up
    volume_id = get_volume_id(instance_id=instance_id)
    if volume_id:
        status = get_volume_status(volume_id)
    else:
        hookenv.log("Cannot find volume name to detach, done")
        return

    if status == "available":
        hookenv.log("Volume (%s) already detached. Done" % volume_id)
        return

    hookenv.log(
        "Detaching volume (%s) from instance %s" % (volume_id, instance_id))
    try:
        subprocess.check_call(
            "nova volume-detach %s %s" % (instance_id, volume_id), shell=True)
    except subprocess.CalledProcessError, e:
        hookenv.log(
            "ERROR: Couldn't detach nova volume %s. %s" % (volume_id, str(e)),
            hookenv.ERROR)
        sys.exit(1)
    return
