from charmhelpers import fetch
import hooks
import mocker
from testing import TestHookenv


class TestHooks(mocker.MockerTestCase):

    def setUp(self):
        self.maxDiff = None
        hooks.hookenv = TestHookenv(
            {"key": "myusername", "tenant": "myusername_project",
             "secret": "password", "region": "region1",
             "endpoint": "https://keystone_url:443/v2.0/",
             "default_volume_size": 11})

    def test_config_changed_missing_required_config_option(self):
        """
        L{config_changed} will exit in error and log a message if required
        configuration options are missing.
        """
        self.addCleanup(
            setattr, hooks.hookenv, "_config", hooks.hookenv._config)
        hooks.hookenv._config = (
            ("key", "myusername"), ("tenant", ""),
            ("secret", "password"), ("region", ""),
            ("endpoint", "https://keystone_url:443/v2.0/"))
        self.mocker.replay()

        result = self.assertRaises(SystemExit, hooks.config_changed)
        self.assertEqual(result.code, 0)
        message = (
            "WARNING: Incomplete charm config. Missing values for: "
            "region,tenant")
        self.assertIn(
            message, hooks.hookenv._log_WARNING, "Not logged- %s" % message)

    def test_config_changed_calls_load_environment_when_config_is_ready(self):
        """
        L{config_changed} will call L{load_environment} to load and validate
        configured credentials when all mandatory configuration options are
        set.
        """
        load_environment = self.mocker.replace("nova_util.load_environment")
        load_environment()
        self.mocker.replay()
        hooks.config_changed()

    def test_install_adds_apt_source_and_installs_novaclient(self):
        """
        L{install} will call C{fetch.add_source} to add a cloud repository and
        install the C{python-novaclient} package.
        """
        add_source = self.mocker.replace(fetch.add_source)
        add_source("cloud-archive:havana")
        apt_update = self.mocker.replace(fetch.apt_update)
        apt_update(fatal=True)
        apt_install = self.mocker.replace(fetch.apt_install)
        apt_install(["python-novaclient"], fatal=True)
        self.mocker.replay()

        hooks.install()

    def test_block_storage_relation_changed_waits_without_instance_id(self):
        """
        L{block_storage_relation_changed} will log a message and wait when
        the relation does not yet contain the required C{instance-id}.
        """
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = ()

        self.assertIsNone(hooks.hookenv.relation_get("instance-id"))
        hooks.block_storage_relation_changed()
        message = "Waiting for relation to define instance-id"
        self.assertIn(
            message, hooks.hookenv._log_INFO, "Not logged- %s" % message)

    def test_block_storage_relation_changed_with_instance_id(self):
        """
        L{block_storage_relation_departed} calls L{attach_nova_volume} when
        C{instance-id} is available in the relation data. To report
        a successful device attach, it sets the relation data
        C{block-device-path} to the attached volume's device path from nova.
        """
        device_path = "/dev/vdc"
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (("instance-id", "i-123"),)

        nova_attach = self.mocker.replace("nova_util.attach_nova_volume")
        nova_attach(
            instance_id="i-123", volume_id=None, size=None, volume_label=None)
        self.mocker.result(device_path)   # The attached device path from nova
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        self.assertIsNone(hooks.hookenv.relation_get("volume-id"))
        self.assertIsNone(hooks.hookenv.relation_get("size"))
        self.assertIsNone(hooks.hookenv.relation_get("volume-label"))
        hooks.block_storage_relation_changed()
        self.assertEqual(
            hooks.hookenv._outgoing_relation_data,
            (("block-device-path", device_path),))

    def test_block_storage_relation_changed_with_instance_id_volume_id(self):
        """
        When C{volume-id} and C{instance-id} are both present in the relation
        data, they will both be passed to L{attach_nova_volume}. To report a
        successful device attach, it sets the relation data
        C{block-device-path} to the attached volume's device path from nova.
        """
        device_path = "/dev/vdc"
        volume_id = "123-123-123"
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (
            ("instance-id", "i-123"), ("volume-id", volume_id))

        nova_attach = self.mocker.replace("nova_util.attach_nova_volume")
        nova_attach(
            instance_id="i-123", volume_id=volume_id, size=None,
            volume_label=None)
        self.mocker.result(device_path)   # The attached device path from nova
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        self.assertEqual(hooks.hookenv.relation_get("volume-id"), volume_id)
        self.assertIsNone(hooks.hookenv.relation_get("size"))
        self.assertIsNone(hooks.hookenv.relation_get("volume-label"))
        hooks.block_storage_relation_changed()
        self.assertEqual(
            hooks.hookenv._outgoing_relation_data,
            (("block-device-path", device_path),))

    def test_block_storage_relation_changed_with_instance_id_size(self):
        """
        When C{size} and C{instance-id} are both present in the relation data,
        they will be passed to L{attach_nova_volume}. To report a successful
        device attach, it sets the relation data C{block-device-path} to the
        attached volume's device path from nova.
        """
        device_path = "/dev/vdc"
        size = 12
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (
            ("instance-id", "i-123"), ("size", size))

        nova_attach = self.mocker.replace("nova_util.attach_nova_volume")
        nova_attach(
            instance_id="i-123", volume_id=None, size=size, volume_label=None)
        self.mocker.result(device_path)   # The attached device path from nova
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        self.assertIsNone(hooks.hookenv.relation_get("volume-id"))
        self.assertEqual(hooks.hookenv.relation_get("size"), size)
        self.assertIsNone(hooks.hookenv.relation_get("volume-label"))
        hooks.block_storage_relation_changed()
        self.assertEqual(
            hooks.hookenv._outgoing_relation_data,
            (("block-device-path", device_path),))

    def test_block_storage_relation_departed_fails_without_instance_id(self):
        """
        L{block_storage_relation_departed} will exit in error and log a message
        when the relation does not contain the required C{instance-id}.
        """
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = ()

        self.assertIsNone(hooks.hookenv.relation_get("instance-id"))
        result = self.assertRaises(
            SystemExit, hooks.block_storage_relation_departed)
        self.assertEqual(result.code, 1)
        message = "Cannot detach nova volume from instance without instance-id"
        self.assertIn(
            message, hooks.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_block_storage_relation_departed_detaches_volume(self):
        """
        L{block_storage_relation_departed} calls L{detach_nova_volume} when
        C{instance-id} is available in the relation data.
        """
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (("instance-id", "i-123"),)

        nova_detach = self.mocker.replace("nova_util.detach_nova_volume")
        nova_detach("i-123")
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        hooks.block_storage_relation_departed()
