import nova_util as util
import mocker
import os
import subprocess
from testing import TestHookenv


class TestNovaUtil(mocker.MockerTestCase):

    def setUp(self):
        self.maxDiff = None
        util.hookenv = TestHookenv(
            {"key": "myusername", "tenant": "myusername_project",
             "secret": "password", "region": "region1",
             "endpoint": "https://keystone_url:443/v2.0/",
             "default_volume_size": 11})
        util.log = util.hookenv.log

    def test_load_environment_with_nova_variables(self):
        """
        L{load_environment} will setup script environment variables for nova
        by mapping configuration values provided to openstack OS_* environment
        variables and then call L{validate_credentials} to assert
        that environment variables provided give access to the service.
        """
        self.addCleanup(setattr, util.os, "environ", util.os.environ)
        util.os.environ = {}
        credentials = self.mocker.replace(util.validate_credentials)
        credentials()
        self.mocker.replay()

        util.load_environment()
        expected = {
            "OS_AUTH_URL": "https://keystone_url:443/v2.0/",
            "OS_PASSWORD": "password",
            "OS_REGION_NAME": "region1",
            "OS_TENANT_NAME": "myusername_project",
            "OS_USERNAME": "myusername"
        }
        self.assertEqual(util.os.environ, expected)

    def test_load_environment_error_missing_config_options(self):
        """
        L{load_environment} will exit in failure and log a message if any
        required configuration option is not set.
        """
        self.addCleanup(setattr, util.os, "environ", util.os.environ)
        credentials = self.mocker.replace(util.validate_credentials)
        credentials()
        self.mocker.throw(SystemExit)
        self.mocker.replay()

        self.assertRaises(SystemExit, util.load_environment)

    def test_validate_credentials_failure(self):
        """
        L{validate_credentials} will attempt a simple nova command to ensure
        the environment is properly configured to access the nova service.
        Upon failure to contact the nova service, L{validate_credentials} will
        exit in error and log a message.
        """
        command = "nova list"
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(SystemExit, util.validate_credentials)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Charm configured credentials can't access endpoint. "
            "Command '%s' returned non-zero exit status 1" % command)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_validate_credentials(self):
        """
        L{validate_credentials} will succeed when a simple nova command
        succeeds due to a properly configured environment based on the charm
        configuration options.
        """
        command = "nova list"
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.replay()

        util.validate_credentials()
        message = (
            "Validated charm configuration credentials have access to "
            "block storage service"
        )
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_get_volume_attachments_present(self):
        """
        L{get_volume_attachments} returns a C{list} of available volume
        attachments for the given C{volume_id}.
        """
        volume_id = "123-123-123"
        command = (
            "nova volume-show %s | grep attachments | awk -F '|' '{print $3}'"
            % volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result(
            "[{u'device': u'/dev/vdc', u'server_id': u'blah', "
            "u'id': u'i-123123', u'volume_id': u'%s'}]" % volume_id)
        self.mocker.replay()

        expected = [{
            "device": "/dev/vdc", "server_id": "blah", "id": "i-123123",
            "volume_id": volume_id}]

        self.assertEqual(util.get_volume_attachments(volume_id), expected)

    def test_get_volume_attachments_no_attachments_present(self):
        """
        L{get_volume_attachments} returns an empty C{list} if no available
        volume attachments are reported for the given C{volume_id}.
        """
        volume_id = "123-123-123"
        command = (
            "nova volume-show %s | grep attachments | awk -F '|' '{print $3}'"
            % volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result("[]")
        self.mocker.replay()

        self.assertEqual(util.get_volume_attachments(volume_id), [])

    def test_get_volume_attachments_no_volume_present(self):
        """
        L{get_volume_attachments} returns an empty C{list} if no available
        volume is discovered for the given C{volume_id}.
        """
        volume_id = "123-123-123"
        command = (
            "nova volume-show %s | grep attachments | awk -F '|' '{print $3}'"
            % volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        self.assertEqual(util.get_volume_attachments(volume_id), [])

    def test_volume_exists_true(self):
        """
        L{volume_exists} returns C{True} when C{volume_id} is seen by the nova
        client command C{nova volume-show}.
        """
        volume_id = "123134124-1241412-1242141"
        command = "nova volume-show %s" % volume_id
        nova_cmd = self.mocker.replace(subprocess.call)
        nova_cmd(command, shell=True)
        self.mocker.result(0)
        self.mocker.replay()
        self.assertTrue(util.volume_exists(volume_id))

    def test_volume_exists_false(self):
        """
        L{volume_exists} returns C{False} when C{volume_id} is not seen by the
        nova client command C{nova volume-show}.
        """
        volume_id = "123134124-1241412-1242141"
        command = "nova volume-show %s" % volume_id
        nova_cmd = self.mocker.replace(subprocess.call)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, "Volume not here"))
        self.mocker.replay()

        self.assertFalse(util.volume_exists(volume_id))

    def test_get_volume_id_by_volume_name(self):
        """
        L{get_volume_id} provided with a existing C{volume_name} returns the
        corresponding nova volume id.
        """
        volume_name = "my-volume"
        volume_id = "12312412-412312\n"
        command = (
            "nova volume-show '%s' | grep ' id ' | awk '{ print $4 }'" %
            volume_name)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result(volume_id)
        self.mocker.replay()
        self.assertEqual(util.get_volume_id(volume_name), volume_id.strip())

    def test_get_volume_id_command_error(self):
        """
        L{get_volume_id} handles any nova command error by reporting the error
        and exiting the hook.
        """
        volume_name = "my-volume"
        command = (
            "nova volume-show '%s' | grep ' id ' | awk '{ print $4 }'" %
            volume_name)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(SystemExit, util.get_volume_id, volume_name)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't find nova volume id for %s. Command '%s' "
            "returned non-zero exit status 1" % (volume_name, command))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_get_volume_id_without_volume_name(self):
        """
        L{get_volume_id} without a provided C{volume_name} will discover the
        nova volume id by searching nova volume-list for volumes labelled with
        the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        self.addCleanup(
            setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        volume_id = "123134124-1241412-1242141\n"
        command = (
            "nova volume-list | grep %s | awk '{print $2}'" % unit_name)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result(volume_id)
        self.mocker.replay()

        self.assertEqual(util.get_volume_id(), volume_id.strip())

    def test_get_volume_id_without_volume_name_no_matching_volume(self):
        """
        L{get_volume_id} without a provided C{volume_name} will return C{None}
        when it cannot find a matching volume label from nova volume-list for
        the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        self.addCleanup(
            setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        command = (
            "nova volume-list | grep %s | awk '{print $2}'" % unit_name)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result("\n")   # Empty result string from awk
        self.mocker.replay()

        self.assertIsNone(util.get_volume_id())

    def test_get_volume_id_without_volume_name_multiple_matching_volumes(self):
        """
        L{get_volume_id} does not support multiple volumes associated with the
        the instance represented by os.environ[JUJU_REMOTE_UNIT]. When
        C{volume_name} is not specified and nova volume-list returns multiple
        results the function exits with an error.
        """
        unit_name = "postgresql/0"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        command = (
            "nova volume-list | grep %s | awk '{print $2}'" % unit_name)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result("123-123-123\n456-456-456\n")   # Two results
        self.mocker.replay()

        result = self.assertRaises(SystemExit, util.get_volume_id)
        self.assertEqual(result.code, 1)
        message = (
            "Error: Multiple nova volumes labeled as associated with "
            "%s. Cannot get_volume_id." % unit_name)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_get_volume_status_by_known_volume_id(self):
        """
        L{get_volume_status} returns the status of a volume matching
        C{volume_id} by using the nova client commands.
        """
        volume_id = "123134124-1241412-1242141"
        command = (
            "nova volume-show '%s' | grep ' status ' | awk '{ print $4 }'" %
            volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result("available\n")
        self.mocker.replay()
        self.assertEqual(util.get_volume_status(volume_id), "available")

    def test_get_volume_status_by_invalid_volume_id(self):
        """
        L{get_volume_status} returns the status of a volume matching
        C{volume_id} by using the nova client commands.
        """
        volume_id = "123134124-1241412-1242141"
        command = (
            "nova volume-show '%s' | grep ' status ' | awk '{ print $4 }'" %
            volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()
        self.assertIsNone(util.get_volume_status(volume_id))
        message = (
            "Error: nova couldn't get status of volume %s. "
            "Command '%s' returned non-zero exit status 1" %
            (volume_id, command))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_get_volume_status_when_get_volume_id_none(self):
        """
        L{get_volume_status} logs a warning and returns C{None} when
        C{volume_id} is not specified and L{get_volume_id} returns C{None}.
        """
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id()
        self.mocker.result(None)
        self.mocker.replay()

        self.assertIsNone(util.get_volume_status())
        message = "WARNING: Can't find volume_id to get status."
        self.assertIn(
            message, util.hookenv._log_WARNING, "Not logged- %s" % message)

    def test_get_volume_status_when_get_volume_id_discovers(self):
        """
        When C{volume_id} is not specified, L{get_volume_status} obtains the
        volume id from L{get_volume_id} gets the status using nova commands.
        """
        volume_id = "123-123-123"
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id()
        self.mocker.result(volume_id)
        command = (
            "nova volume-show '%s' | grep ' status ' | awk '{ print $4 }'" %
            volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result("in-use\n")
        self.mocker.replay()

        self.assertEqual(util.get_volume_status(), "in-use")

    def test_attach_nova_volume_failure_when_volume_id_does_not_exist(self):
        """
        When L{attach_nova_volume} is provided a C{volume_id} that doesn't
        exist it logs and error and exits.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        volume_exists = self.mocker.replace(util.volume_exists)
        volume_exists(volume_id)
        self.mocker.result(False)
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, util.attach_nova_volume, instance_id, volume_id)
        self.assertEqual(result.code, 1)
        message = (
            "Requested volume-id (%s) does not exist. "
            "Unable to associate storage with %s" % (volume_id, unit_name))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_nova_volume_when_volume_id_already_attached(self):
        """
        When L{attach_nova_volume} is provided a C{volume_id} that already
        has the state C{in-use} it logs that the volume is already attached
        and returns.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        volume_exists = self.mocker.replace(util.volume_exists)
        volume_exists(volume_id)
        self.mocker.result(True)
        get_vol_status = self.mocker.replace(util.get_volume_status)
        get_vol_status(volume_id)
        self.mocker.result("in-use")
        get_attachments = self.mocker.replace(util.get_volume_attachments)
        get_attachments(volume_id)
        self.mocker.result([{"device": "/dev/vdc"}])
        self.mocker.replay()

        self.assertEqual(
            util.attach_nova_volume(instance_id, volume_id), "/dev/vdc")

        message = "Volume %s already attached. Done" % volume_id
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_nova_volume_failure_when_volume_unsupported_status(self):
        """
        When L{attach_nova_volume} is provided a C{volume_id} that has an
        unsupported status. It logs the error and exits.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        volume_exists = self.mocker.replace(util.volume_exists)
        volume_exists(volume_id)
        self.mocker.result(True)
        get_vol_status = self.mocker.replace(util.get_volume_status)
        get_vol_status(volume_id)
        self.mocker.result("deleting")
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, util.attach_nova_volume, instance_id, volume_id)
        self.assertEqual(result.code, 1)
        message = ("Cannot attach nova volume. "
                   "Volume has unsupported status: deleting")
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_nova_volume_creates_with_config_size(self):
        """
        When C{volume_id} is C{None}, L{attach_nova_volume} will create a new
        nova volume with the configured C{default_volume_size} when the volume
        doesn't exist and C{size} is not provided.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        volume_label = "%s unit volume" % unit_name
        default_volume_size = util.hookenv.config("default_volume_size")
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id(volume_label)
        self.mocker.result(None)
        command = (
            "nova volume-create --display-name '%s' %s" %
            (volume_label, default_volume_size))
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        get_vol_id(volume_label)
        self.mocker.result(volume_id)   # Found the volume now
        get_vol_status = self.mocker.replace(util.get_volume_status)
        get_vol_status(volume_id)
        self.mocker.result("available")
        command = (
            "nova volume-attach %s %s auto | egrep -o \"/dev/vd[b-z]\"" %
            (instance_id, volume_id))
        attach_cmd = self.mocker.replace(subprocess.check_output)
        attach_cmd(command, shell=True)
        self.mocker.result("/dev/vdc\n")
        self.mocker.replay()

        self.assertEqual(util.attach_nova_volume(instance_id), "/dev/vdc")
        messages = [
            "Creating a %sGig volume named (%s) for instance %s" %
            (default_volume_size, volume_label, instance_id),
            "Attaching %s (%s)" % (volume_label, volume_id)]
        for message in messages:
            self.assertIn(
                message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_nova_volume_no_volume_found(self):
        """
        When L{get_volume_id} is unable to find an attached volume and returns
        C{None}, L{detach_volume} will log a message and perform no work.
        """
        unit_name = "postgresql/0"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        volume_label = "%s unit volume" % unit_name

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id(volume_label)
        self.mocker.result(None)
        self.mocker.replay()

        util.detach_nova_volume("i-123123")
        message = "Cannot find volume name to detach, done"
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_nova_volume_volume_already_detached(self):
        """
        When L{get_volume_id} finds a volume that is already C{available} it
        logs that the volume is already detached and does no work.
        """
        unit_name = "postgresql/0"
        volume_id = "123-123-123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        volume_label = "%s unit volume" % unit_name

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id(volume_label)
        self.mocker.result(volume_id)
        get_vol_status = self.mocker.replace(util.get_volume_status)
        get_vol_status(volume_id)
        self.mocker.result("available")
        self.mocker.replay()

        util.detach_nova_volume("i-123123")  # pass in our instance_id
        message = "Volume (%s) already detached. Done" % volume_label
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_nova_volume_command_error(self):
        """
        When the nova volume-detach command fails, L{detach_nova_volume} will
        log a message and exit in error.
        """
        unit_name = "postgresql/0"
        volume_id = "123-123-123"
        instance_id = "i-123123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        volume_label = "%s unit volume" % unit_name

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id(volume_label)
        self.mocker.result(volume_id)
        get_vol_status = self.mocker.replace(util.get_volume_status)
        get_vol_status(volume_id)
        self.mocker.result("in-use")
        command = "nova volume-detach %s %s" % (instance_id, volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, util.detach_nova_volume, instance_id)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't detach nova volume %s. Command '%s' "
            "returned non-zero exit status 1" % (volume_id, command))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_detach_nova_volume(self):
        """
        When L{get_volume_id} finds a volume associated with this instance
        which has a volume state not equal to C{available}, it detaches that
        volume using nova commands.
        """
        unit_name = "postgresql/0"
        volume_id = "123-123-123"
        instance_id = "i-123123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": unit_name}
        volume_label = "%s unit volume" % unit_name

        load_environment = self.mocker.replace(util.load_environment)
        load_environment()
        get_vol_id = self.mocker.replace(util.get_volume_id)
        get_vol_id(volume_label)
        self.mocker.result(volume_id)
        get_vol_status = self.mocker.replace(util.get_volume_status)
        get_vol_status(volume_id)
        self.mocker.result("in-use")
        command = "nova volume-detach %s %s" % (instance_id, volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.replay()

        util.detach_nova_volume(instance_id)
        message = (
            "Detaching volume %s (%s) from instance %s" %
            (volume_label, volume_id, instance_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)
